/* ***** BEGIN LICENSE BLOCK *****
 * Licensed under Version: MPL 1.1/GPL 2.0/LGPL 2.1
 * Full Terms at http://mozile.mozdev.org/license.html
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Playsophy code (www.playsophy.com).
 *
 * The Initial Developer of the Original Code is Playsophy
 * Portions created by the Initial Developer are Copyright (C) 2002-2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * 
 * Last revised by Richard Prescott
 * Last revised on 2004-04-18
 * Added some namespace support
 * 
 *
 * ***** END LICENSE BLOCK ***** */

/**
@fileoverview

eDOMXHTML.js

<p>The basic w3c DOM is not honed for interactive editing. eDOM extends the standard DOM
to make it easy to build CSS-enabled editing applications that work
within a browser. eDOMXHTML builds on eDOM for XHTML specific routines.

<p>This file represents the limits to completely CSS and Validation Module based
editing. You need to have HTML specific logic in some cases. As Mozile and eDOM
evolve this file will expand and contract. It is preferable if it goes away but
that may not be possible or efficient.

<p>POST04: 
<ul>
<li>move all XHTML specific routines in here. This will include Range routines
 (part of split of eDOM into eDOMCSS, eDOMXHTML) in particular containerless
 line stuff that mandates inserting "div" parents
<li>open question: is new validation module useful for XHTML editing?
</ul> 

<h3>xhtml lists</h3>

<p>These methods are XHTML specific (there's a general problem with lists
as CSS has no "list-container" display type) and partially selection model specific.

<p>To Do: Nested lists, Token handling for Nested Lists.

<p>POST04:
<ul>
<li>turn methods into those of an XHTML Range object.
<li>this stuff could move into eDOM iff methods allow a user to pass in array of 
 list container names. Consider this step but only after work on issues with DL.
<li>add support for definition list: dl/dt/dd methods
</ul>
*/

/**
Constructor function for eDOMXHTML properties and methods.
Access these via the global object <code>eDOMXHTML</code>.

<p>Note: Many of these methods could move to an XHTML Range object, or possibly
into eDOM if made universal.
*/
function EDOMXHTML(){
};
var eDOMXHTML = new EDOMXHTML();

/**
XHTML Namespace.
@final
*/
EDOMXHTML.prototype.NS = "http://www.w3.org/1999/xhtml";

/**
Default Container Name, "div".
Sometimes bounded or list-item contained lines need to be converted into default contained lines.
The default contained line could be "div", "p", or something else instead.
At the moment it is a single value, but it could be changed to insert a context sensitive container.

<p>The value can be changed via the page MES.
*/
EDOMXHTML.prototype.defaultContainer = "div";


// Check at startup if the MES contains an override for the default container. 
if ("mozileMES" in document && document.mozileMES.eDOM.xhtmlDefaultContainer ) {
	eDOMXHTML.defaultContainer = document.mozileMES.eDOM.xhtmlDefaultContainer;
}

/**
Is HTML/XHTML of empty content type?

<p>area, base, basefont, br, col, frame, hr, img, input, isindex, link, meta, param, object.

@return true if XHTML non child bearing element.
@type boolean
*/
Element.prototype.isXHTMLEmptyContentType = function ()
{
	var elementName = this.XHTMLElementName();
	if (elementName) {
		switch(elementName) {
			case "area":
			case "base":
			case "basefont":
			case "br":
			case "col":
			case "frame":
			case "hr":
			case "img":
			case "input":
			case "isindex":
			case "link":
			case "meta":
			case "param": 
			case "object": // Issue: is allowed but not handled well so disallow
				return true;
		}
	}
	return false;
}

Element.prototype.isXHTMLHeading = function() {
	if ( this instanceof HTMLHeadingElement ) {
		return true;
	} else {
		return false;
	}
}

/**
mozCE: Gets the XHTML ancestor link node &lt;a&gt; for this (text) node.

@return an ancestor link node, or null if there isn't one.
@type Element
@private 
*/
Node.prototype.getLinkAncestor = function() {
	var n = this;
	var a = null;
	while( n.parentNode ) {
		if (n.nodeType==Node.ELEMENT_NODE) {
			if (!n.userModifiable) {
				break;
			} else if (n instanceof HTMLAnchorElement) {
				a = n;
				break;
			} else if (n.computedStyleValue("display") == "block") {
				break;
			}
		}
		n = n.parentNode;
	}
	return a;
}


/**
Gets the ancestor list container for <code>node</code> with a nodeName of "ul", "UL", "ol", or "OL".
This could be <code>node</code> itself.

<p>To Do:<br />
Namespace and text/html issues.<br />
Could have a switch to allow for stricter checking: e.g. parentNode only.  

@param {Node} node first node to start search from.
@return listContainer - list container, or null if not found.
@type Element
@private
*/
EDOMXHTML.prototype.listContainer = function (node)
{
	var n = node;
	while (n != document.documentElement ) {
		if (n.nodeType == Node.ELEMENT_NODE) {
			switch (n.XHTMLElementName()) {
				case "ul" : case "ol" :
					return n;
			}
		}
		n = n.parentNode;
	}
	return null;	
}	

/**
Toggle ul/ol list lines in the selected range.
<ul>
<li>Contained and Bounded Lines that are not already list lines are changed to list-items.
<li>List Containers of the wrong sort are changed. i.e. ul to ol, or ol to ul.
<li>Adjacent lists of the same type, provided they are in the selected range, are merged.
<li>If there are only list item lines of type listContainerName in the range the
list-items are outdented one level. Adjacent lists are NOT merged.
</ul>

@param {CSSRange} cssr CSS Range to toggle.
@param {String} listContainerName "ul" or "ol"
@param {String} listContainerToChange "ol" or "ul"
*/
EDOMXHTML.prototype.listLinesToggle = function (cssr, listContainerName, listContainerToChange)
{
 try{
	listContainerName = listContainerName.toLowerCase();
	listContainerToChange = listContainerToChange.toLowerCase();
	if ( !((listContainerName=="ol" && listContainerToChange=="ul") || (listContainerName=="ul" && listContainerToChange=="ol")) )
		return;

	var listItemAncestor, listContainer, newContainer, i, range;
	var newListElement = false;
	var toggleCount=0;
	var topToNormalize = cssr.commonAncestorContainer.parentElement;
	var lines = cssr.lines;

	//First check for any existing list lines.
	for(i=0; i<lines.length; i++) {
		listItemAncestor = lines[i].listItemAncestor;
		if ( listItemAncestor ) {
			listContainer = eDOMXHTML.listContainer(listItemAncestor);
			if((listItemAncestor != cssr.top) && (listContainer != cssr.top)) {
				if( listContainer.nodeNamed(listContainerToChange) ) {
					// Change any list container names that need to be changed.
					toggleCount++;
					newContainer = eDOMXHTML.createElement(listContainerName);
					newContainer = listContainer.parentNode.insertBefore(newContainer, listContainer);
					while (listContainer.hasChildNodes())
						newContainer.appendChild(listContainer.firstChild);
					listContainer.parentNode.removeChild(listContainer);
				}
			}
		}
	}

	//Convert any Bounded lines that do not have a list-item ancestor to Contained Lines first.
	//This prevents Bounded lines from losing their start container if the start container
	//is a Contained line (e.g. "p")and gets converted to a list line "li". 
	for(i=0; i<lines.length; i++) {
		if(lines[i].lineType == CSSLine.BOUNDED_LINE && !(lines[i].listItemAncestor) ) {
			lines[i] = lines[i].setContainer(eDOMXHTML.createElement("li"), true);
			lines[i].normalizeWhitespace();
			if(lines[i].lineType == CSSLine.CONTAINED_LINE) {
				newContainer = eDOMXHTML.createElement(listContainerName);
				lines[i].container.parentNode.insertBefore( newContainer, lines[i].container );
				newContainer.appendChild( lines[i].container );
				newListElement = true; 
			}				
		}
	}
	
	//Now convert all Contained Lines that do not already have a list-item ancestor.
	for(i=0; i<lines.length; i++) {
		if(!lines[i].listItemAncestor) {
			if(lines[i].lineType == CSSLine.CONTAINED_LINE) {
				//Set container for line. (won't if top)
				lines[i] = lines[i].setContainer(eDOMXHTML.createElement("li"), true);
				lines[i].normalizeWhitespace();
				newContainer = eDOMXHTML.createElement(listContainerName);
				lines[i].container.parentNode.insertBefore( newContainer, lines[i].container );
				newContainer.appendChild( lines[i].container );
				newListElement = true; 
				if( topToNormalize.descendent(lines[i].container.parentNode) ) {
					topToNormalize = lines[i].container.parentNode; //MAX - does this go high enough up or should it be .parentNode again??? 
				}
			}				
		}
	}

	if(newListElement || toggleCount>0 ) {
		var firstLine=null;
		var lastLine=null;
		for (i=0; i<lines.length; i++) {
			if (lines[i].lineType==CSSLine.CONTAINED_LINE) {
				if(firstLine==null)
					firstLine = i;
				lastLine = i;
			}
		}
		if (firstLine==null) {
			firstLine=0;
			lastLine=lines.length-1;
		}
		var rangeStart = lines[firstLine].topMostContainer;
		var rangeEnd = lines[lastLine].topMostContainer;
		range = document.createRange();
		range.setStartBefore(rangeStart);
		range.setEndAfter(rangeEnd);
		range.normalizeElements(listContainerName);
	} else { //if( !newListElement && toggleCount==0 ) {
		eDOMXHTML.outdentLists(cssr);
	}
 }catch(e){
 }
}

/**
Indents Lists in the selected range.

<p>Lists are nested with the addition of editing tokens as required.
Only the top most list-items within each selected list are indented.
There is no limit to the number of indents.

<p>E.g. after indenting lines two and three of a list we would have something like this;

<pre>
&lt;ul&gt;
  &lt;li&gt;one&lt;/li&gt;
  &lt;li&gt;
    &amp;nbsp;
    &lt;ul&gt;
      &lt;li&gt;two&lt;/li&gt;
      &lt;li&gt;three&lt;/li&gt;
    &lt;/ul&gt;
  &lt;/li&gt;
  &lt;li&gt;four&lt;/li&gt;
&lt;/ul&gt;
</pre>

@param {CSSRange} cssr selected range to indent.
*/
EDOMXHTML.prototype.indentLists = function(cssr)
{
 try{
	var lines = cssr.lines;	

	//Split list lines from non-list lines.
	var listElements = new Array();

	var i, j;
	for(i=0; i<lines.length; i++) {
		var listElement = lines[i].listItemAncestor;
		if (!listElement) continue;
		var listContainer = eDOMXHTML.listContainer(listElement);
		if (!listContainer) continue;
		if ( listElement.descendent(lines[i].top) ) {
			if ( listElements.length == 0 || !listContainer.descendent(listElements[listElements.length-1][0]) ) {
				//Get an array of selected list containers, each list container has an array of list-items.
				//Each list container and list item is unique, each listElement is in the editable region.
				//Nested list containers are discarded, only the top most list items are indented. 
				for(j=0; j<listElements.length; j++) {
					if (listElements[j][0] == listContainer )
						break;
				}
				if ( j==listElements.length ) {
				//if new container, add container and list item element
					listElements[j] = new Array();
					listElements[j][0] = listContainer;
					listElements[j][1] = new Array();
					listElements[j][1].push(listElement);
				} else {
					if (listElement != listElements[j][1][listElements[j][1].length-1])
						listElements[j][1].push(listElement);
				}
			}
		}
	}
	if(listElements.length > 0) {
		// indent the list elements
		for(i=0; i<listElements.length; i++) {
			listContainer = listElements[i][0];
			var topElement = listElements[i][1][0];

			//Create the new sub list: e.g. <li>nbsp<ul></ul>nbsp</li>
			var newListItem = eDOMXHTML.createElement("li");
			newListItem.appendChild(eDOM.token());
			var newListContainer = newListItem.appendChild(listContainer.cloneNode(false));
//			newListItem.appendChild(eDOM.token());

			//Insert the new empty sublist
			topElement.parentNode.insertBefore( newListItem, topElement );
			
			//Now append all the selected list items
			for (j=0; j<listElements[i][1].length; j++ ) {
				newListContainer.appendChild(listElements[i][1][j]);
			}
		}
		//Update range pointers.
		cssr.__reMarkTextBoundaries();
	}
 }catch(e){
 }
}
	
/**
Outdent (nested) lists; "ol" or "ul" in the selected range.
Only the top most list-items within each selected list are outdented.
Top level list-item lines are enclosed with the xhtml default container if they
do not have their own container.

@param {CSSRange} cssr selected CSS Range to outdent.
*/
EDOMXHTML.prototype.outdentLists = function(cssr)
{
 try{
	var lines = cssr.lines;

	var listElements = new Array();

	var i, j;
	for(i=0; i<lines.length; i++) {
		var listElement = lines[i].listItemAncestor;
		if (!listElement) continue;
		var listContainer = eDOMXHTML.listContainer(listElement);
		if (!listContainer) continue;
		if ( listContainer.descendent(lines[i].top) ) {
			if ( listElements.length == 0 || !listContainer.descendent(listElements[listElements.length-1][0]) ) {
				//Get an array of selected list containers, each list container has an array of list-items.
				//Each list container and list item is unique, each listContainer is in the editable region.
				//Nested list containers are discarded, only the top most container is outdented. 

				//First search list to see if container is already in the list.
				for(j=0; j<listElements.length; j++) {
					if (listElements[j][0] == listContainer )
						break;
				}
				if ( j==listElements.length ) {
				//if new container, add container and list item element
					listElements[j] = new Array();
					listElements[j][0] = listContainer;
					listElements[j][1] = new Array();
					listElements[j][1].push(listElement);
				} else {
				//Add list item element, provided it is not already present.
					if (listElement != listElements[j][1][listElements[j][1].length-1])
						listElements[j][1].push(listElement);
				}
			}
		}
	}

	if(listElements.length > 0) {
		// Outdent the list elements
		for(i=0; i<listElements.length; i++) {
			for(j=0; j<listElements[i][1].length; j++ ) {
				eDOMXHTML.__outdentListItem(listElements[i][0], listElements[i][1][j]);
			}
		}
	}
 }catch(e){
 }
}

/**
Outdents a single XHTML List Item.

<p>To Do:<br />
Still work to do on tidying up empty token lines.
Account for other (block) content at the same level as the nested list container.

@param {Element} listContainer
@param {Element} listItem
@private
*/
EDOMXHTML.prototype.__outdentListItem = function(listContainer, listItem)
{
	// if not the first element in the container then split the container and work on the new container
	var newContainer = listContainer.split(listItem.offset);
	var parentLi = null;
	if(newContainer) {
		listItem = newContainer.firstChild;
		//If a nested list, append the split container to a clone of the parent "li" node.
		if ( newContainer.parentNode.computedStyleValue("display") == "list-item" ) {
			parentLi = newContainer.parentNode.cloneNode(false);
			parentLi = newContainer.parentNode.parentNode.insertAfter( parentLi, newContainer.parentNode ); 
			parentLi.appendChild( newContainer );
		}
	}

	var container = listItem.parentNode;
	if ( !parentLi && container.parentNode.computedStyleValue("display") == "list-item" ) {
		//Don't forget about the first item in a list, it won't have led to a split.
		parentLi = container.parentNode;
	}
	
	//Promote list item.
	if (parentLi){
		listItem = parentLi.parentNode.insertBefore(listItem, parentLi);
	}else{
		listItem = container.parentNode.insertBefore(listItem, container);
	}

	//Tidy up container;
	if( container.editableFirstChild() != null ) {
		container.normalize();
		if ( parentLi ) {
			var token = document.createTextNode(String.fromCharCode(eDOM.TOKEN));
			parentLi.insertBefore( token , parentLi.firstChild );
		}
	} else {
	//Else listContainer is now empty delete it.
		container.parentNode.removeChild(container);
		//If parent list-item is also empty delete it as well.
		if ( parentLi ) {
			if ( parentLi.editableFirstChild() == null ) {
				parentLi.parentNode.removeChild(parentLi);
			}
		}
	}

	var listAncestor = eDOMXHTML.listContainer(listItem);
	if ( listAncestor == null ) {
	//If listItem is no longer a descendent of a list container;
		// go through lines promoting where necessary
		var listContents = document.createRange();
		listContents.selectNodeContents(listItem);
		listContents = documentCreateCSSTextRange(listContents, listItem.parentNode);
		var linesInList = listContents.lines;
	
		if(linesInList.length == 1) {
		// typical case: only one line
			if(linesInList[0].container == listItem) {
			//If a "li" container only, change "li" to default xhtml container.
				linesInList[0].setContainer(eDOMXHTML.createElement(eDOMXHTML.defaultContainer), true);
			} else {
			//Else the "li" contains another block level container: remove "li" container leaving the promoted contents.
				listItem.parentNode.removeChildOnly(listItem);
			}
		} else {
		// first or last lines bounded - make into contained if necessary
			var l = linesInList[0];
			if((l.lineType == CSSLine.BOUNDED_LINE) && !l.startBoundary && !l.emptyLine) {
				l = l.setContainer(eDOMXHTML.createElement(eDOMXHTML.defaultContainer), true);
				l.normalizeWhitespace();
			}
			l = linesInList[linesInList.length-1];
			if((l.lineType == CSSLine.BOUNDED_LINE) && !l.endBoundary && !l.emptyLine) {
				l = l.setContainer(eDOMXHTML.createElement(eDOMXHTML.defaultContainer), true);
				l.normalizeWhitespace();
			}

			//Remove the "li" container leaving the promoted contents.
			listItem.parentNode.removeChildOnly(listItem);
		}
	}
}


/**
Indent lines ; Sets CSS property "margin-left".

<ul>
<li>Bounded Lines are converted to Contained Lines if possible before indenting.
<li>List Containers (ul and ol) are indented.
<li>TOP and TABLE_CELL Contained Lines are not indented. 
</ul>

@param {CSSRange} cssr selected range to indent.
@param {String} indent CSS value for indent.
*/
EDOMXHTML.prototype.indentLines = function(cssr, indent) {
 try{
	var lines = cssr.lines;
	var containers = new Array();
	
	for(var i=0; i<lines.length; i++) {
		var l = lines[i];
		if (l.lineType == CSSLine.BOUNDED_LINE) {
			if(l.emptyLine && !l.tokenLine )
				continue; //ignore empty line.
			// try and turn bounded line into contained line
			l = l.setContainer(eDOMXHTML.createElement(eDOMXHTML.defaultContainer), true);
			l.normalizeWhitespace();
		}
		if (l.lineType == CSSLine.CONTAINED_LINE) {
			switch (l.containedLineType) {
				case ContainedLine.TOP :
				case ContainedLine.TABLE_CELL :
					continue;
				case ContainedLine.LIST_ITEM :
					var listContainer = eDOMXHTML.listContainer( l.container );
					if (!listContainer) continue; //Will only happen if invalid xhtml!
					for( var c=0; c<containers.length; c++ ){
						if (containers[c] == listContainer)
							break;
					}
					if (c == containers.length) {
					//if container has not already been indented
						containers.push(listContainer);
						if ( listContainer.descendent(l.top) )
						//If not top (i.e. should be editable)
							listContainer.setStyle("margin-left", indent);
					}
					break;
				case ContainedLine.BLOCK :
					l.setStyle("margin-left", indent);
			}
		}
	}
 }catch(e){
 }
}

/********************************* XHTML specific line handling *******************/
/**
eDOMXHTML: insert an xhtml break; "br".

This causes a problem because the browser doesn't seem to select the correct point
at which to show the cursor when traversing through break nodes.

This to be finished...
E.g. Need to add token(?) when at end of line.
+ then need to overwrite token when type next to it.
+ have to delete token and br as one.in one go if nothing beyond Token
*/
InsertionPoint.prototype.breakXHTMLLine = function()
{
	try {
		var line = this.line;
		var textnode = eDOM.newLine();
		var backOne = false; 
		if ( this.endOfLine ) {
			textnode.appendData( String.fromCharCode(eDOM.TOKEN) );
			backOne = true;
		}
		this.insertNode( eDOMXHTML.createElement("br") );
		this.insertNode( textnode );
		line.normalizeTextNodes();
		if ( backOne ) {
			this.backOne();
		}
	} catch (e) {
		alert ( "breakXHTMLline: "+e );
	}
}

/**
eDOMXHTML: Split XHTML line.

<p>This is specific because we may have to explicitly choose the container
element to insert.

@param {Boolean} alwaysSplit true = split a pre section; false = insert newline in pre section. 
*/
InsertionPoint.prototype.splitXHTMLLine = function( alwaysSplit )
{
	var line = this.line;
	var lineOffset, ip, container;
	
	if ( !alwaysSplit && this.isPre() ) {
	//If "pre" then need to insert a linefeed (newline) char.
		this.insertCharacter( eDOM.LINEFEED );
	}
	else if(line.lineType == CSSLine.BOUNDED_LINE) {
	//Need to split Bounded line.	
		// If empty, non-Token line, set to Token Bounded line.
		// (Can this ever occur? Leave it in for now.)
		if(line.emptyLine && !line.tokenLine) {
			this.set(line.setToTokenLine());
			line = this.line;
		}
		// Make bounded line into a contained line.
		lineOffset = this.lineOffset;
		container = eDOMXHTML.createElement(eDOMXHTML.defaultContainer);
		if ( line.lineParent.canAppendChild( container ) !== false ) {
			line = line.setContainer(container, true);
			line.normalizeWhitespace();
			ip = line.insertionPointAt(lineOffset);
			this.set(ip);
			this.splitContainedLine();
		}
	}
	//Else CONTAINED_LINE
	else {
		var clType = line.containedLineType;
		if( clType == ContainedLine.TOP ) {
		// If TOP
			lineOffset = this.lineOffset;
			container = eDOMXHTML.createElement(eDOMXHTML.defaultContainer);
			if ( line.top.canAppendChild( container ) !== false ) {
			// IF .top allows default block container as a child.
				line = line.setContainer(container, false);
				line.normalizeWhitespace();
				ip = line.insertionPointAt(lineOffset);
				this.set(ip);
				this.splitContainedLine();
			}
		}
		else if ( clType == ContainedLine.TABLE_CELL ) {
			lineOffset = this.lineOffset;
			container = eDOMXHTML.createElement(eDOMXHTML.defaultContainer);
			if ( line.container.canAppendChild( container ) !== false ) {
				line = line.setContainer(container, false);
				line.normalizeWhitespace();
				ip = line.insertionPointAt(lineOffset);
				this.set(ip);
				this.splitContainedLine();
			}
		}
		else if ( line.containedLineType == ContainedLine.LIST_ITEM ) {
			this.splitXHTMLList( line );
		}
		else if ( line.containedLineType == ContainedLine.BLOCK ) {
			if ( line.container.parentNode.canAppendChild( line.container ) !== false ) {
				var endOfH = false;
				if ( (line.container instanceof HTMLHeadingElement) && this.endOfLine ) {
					endOfH = true;
				}
				this.splitContainedLine();
				if ( endOfH ) {
					line = this.line;
					line = line.setContainer(eDOMXHTML.createElement("p"), true);
					this.set( line.firstInsertionPoint );
				}
			}
		}
	}
}

/**
Splits a list line.
<p>A list line is split at this Insertion Point into two list lines.
If this list line is the last line in the list, and it is a Token Line,
it is deleted and a new default token line is added after the list container. 

<p>Updates this Insertion Point appropriately to point to the first IP after
the split point.

@param {CSSContainedLine} line list line to split.
@private
*/
InsertionPoint.prototype.splitXHTMLList = function( line )
{
	if ( line.tokenLine ) {
		var listItem = line.container;
		var endItem = true;
		for ( var n = listItem.nextSibling; n != null; n = n.nextSibling ) {
			if(n.computedStyleValue("display") == "list-item") {
				endItem = false;
				break;
			}
		}
	
		if ( endItem ) {
			var listContainer = eDOMXHTML.listContainer( listItem );
			var nextLine = line.nextLine;
			if ( listContainer && listContainer.descendent( line.top ) ) {
				line.deleteLine();
				var newline = eDOMXHTML.createElement(eDOMXHTML.defaultContainer);
				newline.appendChild( eDOM.token() );
				listContainer.parentNode.insertAfter( newline, listContainer );
				var ip = newline.firstInsertionPoint( this.top );
				this.set( ip );
			}
		} else {
			this.splitContainedLine();
		}
	} else {
		this.splitContainedLine();
	}
}
/******************************** XHTML namespace ******************************

Ensure that element creation uses the XHTML name space when required.

For text/html documents namespace is probably not required.

For XHTML specific documents (application/xhtml+xml) the XHTML namespace should
be used. Maybe this should be read from the document rather than being
hard coded here? It is unlikely, but XHTML documents may also contain
other namespaces. e.g. MathML.

For xhtml embedded in XML documents the XHTML namespace should be used for
xhtml elements, and appropriate namespaces for other elements.
May need to add the XHTML namespace declaration at the top of the document
if it is not already present.

When namespaces are being used also need to check for prefix.
i.e. Need to check locaName and not just nodeName
e.g. max:p could be an xhtml "p" element if prefix "max" corresponds to the
XHTML namespace.

Currently the XML support is poor. 

*******************************************************************************/
/**
eDOMXHTML: Gets the xhtml element name, or null if not an xhtml element.
For HTML documents the name is returned in lower case.

This method may need expanding to take into account other cases.

@return name Xhtml Element name or null if not an Xhtml Element.
@type String
*/

Element.prototype.XHTMLElementName = function()
{
	var name = null;
	if ( (document instanceof HTMLDocument) && document.contentType == "text/html" ) {
		name = this.nodeName.toLowerCase();
	}
	else if ( (document instanceof HTMLDocument) && document.contentType == "application/xhtml+xml" ) {
		if (this.namespaceURI == eDOMXHTML.NS)
			name = this.localName;
	}
	else if ( document instanceof XMLDocument ) {
		if (this.namespaceURI == eDOMXHTML.NS)
			name = this.localName;
	}
	return name;
}

/**
Creates an XHTML element.
<p>The XHTML Namespace is used in the following cases;
<ul>
<li>content-type = application/xhtml+xml
<li>XML document
</ul>
<p>The element is created without a namespace for;
<ul>
<li>content-type = text/html
<li>Other sorts of document...
</ul>

<p>To Do: A generalised function that will look at the available evidence
and decide on what namespace, if any, to use.

@param {String} name xhtml element name
@return Element or null if creation failed.
@type Element
*/
EDOMXHTML.prototype.createElement = function( name )
{
	var el;
	try {
		if ( document instanceof HTMLDocument ) {
			if ( document.contentType == "application/xhtml+xml" )
				el = document.createElementNS(eDOMXHTML.NS, name.toLowerCase())
			else // "text/html"
				el = document.createElement(name);
		} else if ( document instanceof XMLDocument ) {
				el = document.createElementNS(eDOMXHTML.NS, name.toLowerCase());
		} else {//your guess is as good as mine as to what this document is...
				el = document.createElement(name);
		}
	} catch (e) {
		el = null;
	}	
	return el;
}

/**
Creates an element with the specified namespace.
Uses EDOMXHTML.createElement if namespace is null.

@param {String} namespace
@param {String} name element name
@return Element or null if creation failed.
@type Element
*/
EDOMXHTML.prototype.createElementNS = function( namespace, name )
{
	var el;
	if (!namespace) {
		el = this.createElement(name);
	} else {
		try {
			el = document.createElementNS(namespace, name);
		} catch (e) {
			el = null;
		}
	}
	return el;
}


function documentCreateXHTMLElementNS(namespace, elementName) {
	if(namespace == null || namespace == "")
		return document.createElement(elementName);
	return document.createElementNS(namespace, elementName);
}

Element.prototype.hasXHTMLAttribute = function(attributeName) {
	if(document.body)
		return this.hasAttribute(attributeName);
	return this.hasAttributeNS(eDOMXHTML.NS, attributeName);
}

Element.prototype.hasXHTMLAttributeNS = function(namespace, attributeName) {
	if(namespace == null || namespace == "")
		return this.hasAttribute(attributeName);
	return this.hasAttributeNS(namespace, attributeName);
}

Element.prototype.getXHTMLAttribute = function(attributeName) {
	if(document.body)
		return this.getAttribute(attributeName);
	return this.getAttributeNS(eDOMXHTML.NS, attributeName);
}

Element.prototype.getXHTMLAttributeNS = function(namespace, attributeName) {
	if(namespace == null || namespace == "")
		return this.getAttribute(attributeName);
	return this.getAttributeNS(namespace, attributeName);
}

Element.prototype.isXHTMLNamed = function(nodeName) {
	if(document.body || this.namespaceURI == eDOMXHTML.NS)
		return(this.nodeName.toLowerCase() == nodeName.toLowerCase());
	return false;
}

Element.prototype.isXHTMLNamedNS = function(namespace, nodeName) {
	if(document.body || this.namespaceURI == eDOMXHTML.NS)
		return(this.nodeName.toLowerCase() == nodeName.toLowerCase());
	return this.nodeName == nodeName && this.namespaceURI == namespace;
}

/********************************** XHTML CSS Range *********************************/

/**
eDOMXHTML : Apply CSS Style to this range.

<p>Assume (for now) that there are no elements partially embedded in elements.<br />
e.g. [span]XX[span]YY[/span][/span]

<p>POST04/05...whenever:
<ul> 
<li>dont' use textNodes ... instead walk tree ala normalize; could use if text node partially in range then
 split ...
<li>consider replacing compareRanges with compareNode (Mozilla): need to move from looking at node contents
 and then including elements that contain those contents to making Range inclusive enough that nodes themselves
 can be checked (efficiency from a comprehensive set of Range boundary setters)
<li>account for top where we won't style the whole of top: use span instead to allow more nuance with subsequent
 settings: insert div in between
<li>allow "inline" span element to be specified explcitly
<li>allow to pass in a number of "inline style holders". For HTML, A and span should be equivalent in that
 A should be split just like span is split as style is applied. STRONG/EM etc must be accounted for too.
 They should be treated as inline style holders within reason. If inline holder has style of some sort
 that needs to be preserved. May also treat FONT/B/I and other deprecated inline style holders as peers
 of span.
<li>recognize non text styles and exception. Some styles should only be applied to elements explicitly 
<li>issue of [a] getting inline style. Not reset - only span's are! They are just overridden.
<li>recognize equivalency between different ways to express the same style. For example,
 "bold" and its numeric equivalent for "font-weight". This may also require accepting a range
 of "equivalent values" for a style that may not be exact matches in official style but which look
 the same. "bolder" and "bold" for instance display the same: setting one where another already 
 applies in a higher context is redundant.
<li>test code needed for messy HTML like spans within or overlapping other spans. The latter isn't 
 allowed in XHTML but most browsers tolerate it.
<li>ensure span is inline before removing it. If it is not inline then it has another purpose.
<li>if applied to text outside the XHTML name space then make sure span has the correct name space
 setting.
<li>support concept of +/-# for numeric settings where settings are applied incrementally when a 
 setting is already present either explicitly or inherited. Used for position and even for making
 font-weight bolder ...
<li>if <span> expands to cover <div> or other parent then merge its style into that parent. Right
 now, certain sequences will leave the span around if style is built up to incrementally cover a
 larger block.
</ul>

@param {string} styleName: CSS style name.
@param {String} styleValue: CSS style value.
*/
Range.prototype.styleText = function(styleName, styleValue)
{
	var styleHolder;
	var siblingStyleHolder;
	var textNodes;

	//If range not selected don't style anything.
	if(this.collapsed) {
		return;
	}

	textNodes = this.textNodes();
	for(var i=0; i<textNodes.length; i++)
	{
		var textContainer = textNodes[i].parentNode;
		
		if( !textContainer.descendent( this.__top ) ) {
			if ( textContainer == this.top ) {
			//If textContainer is "top";
				styleHolder = eDOMXHTML.createElement('span');
				if ( textContainer.canAppendChild( styleHolder ) ) {
				//If textContainer can contain a span;
					//wrap text node in a "span", which then becomes the container.
					textContainer.insertBefore(styleHolder, textNodes[i]);
					styleHolder.appendChild(textNodes[i]);
					textNodes[i] = styleHolder.firstChild;
					textContainer = styleHolder;
				} else {
				//Else can't add a span, so unable to style.
					return; //Exit.
				}
			} else {
			//Else "top" is a descendent of textContainer. (Shouldn't happen).
				return;
			}
		}
		
		if(document.defaultView.getComputedStyle(textContainer, null).getPropertyValue(styleName) != styleValue) {
			//If textContainer doesn't already have the style;	
			if(textContainer.childNodes.length > 1) {
			//If text doesn't have an exclusive parent then we need to give it one.
				// spans are special: we don't embed spans in a span - we put spans around all the text nodes in the span.
				// note: assume not span within a span so we only have a series of text nodes
				if(textContainer.isXHTMLNamed("span") || textContainer.isXHTMLNamed("a")) { 
					if(textNodes[i].previousSibling) {
						siblingStyleHolder = textContainer.cloneNode(false);
						textContainer.parentNode.insertBefore(siblingStyleHolder, textContainer);
						siblingStyleHolder.appendChild(textNodes[i].previousSibling);	
					}
					if(textNodes[i].nextSibling) {
						siblingStyleHolder = textContainer.cloneNode(false);
						if(textContainer.nextSibling)
							textContainer.parentNode.insertBefore(siblingStyleHolder, textContainer.nextSibling);
						else 
							textContainer.parentNode.appendChild(siblingStyleHolder);
						siblingStyleHolder.appendChild(textNodes[i].nextSibling);	
					}									
				}
				// one text node within a non span element - put this text node within a span
				else {
					styleHolder = eDOMXHTML.createElement('span');
					textContainer.insertBefore(styleHolder, textNodes[i]);
					styleHolder.appendChild(textNodes[i]);
					textNodes[i] = styleHolder.firstChild;
					textContainer = styleHolder;
				}
			}
			textContainer.style.setProperty(styleName, styleValue, "");
		}
	}

	this.__restoreTextBoundaries(); // restore boundaries after styles are applied: CAN NIX IF GET CONTAINER AT START?

	var treeTop = this.commonAncestorContainer.topInlineAncestor.parentElement; //Is there a chance that this might include non-editable nodes???
	treeTop.normalize(); // make sure text nodes are normalized
	treeTop.__normalizeXHTMLTextStyle(); // normalize the styles

	this.__restoreTextBoundaries(); // restore boundaries after styles are normalized.
	return;
}

/**
eDOMXHTML : Wraps the text within this range with "elementName".

"span" is replaced by "elementName".
"elementName" is wrapped around "a".
For all other elements "elementName" is inserted into the element.

To Do: Check that it is a valid nesting.

@param {String} elementName
*/
Range.prototype.wrapText = function(elementName)
{
	this.wrapTextNS(eDOMXHTML.NS, elementName);
}

Range.prototype.wrapTextNS = function(namespace, elementName)
{
	// if range is collapsed do nothing.
	if(this.collapsed)
		return;

	// go through all text nodes in the range and split "span" or "a" where required.
	var textNodes = this.textNodes();
	for(var i=0; i<textNodes.length; i++)
	{
		var textContainer = textNodes[i].parentNode;

		// If the selected text is part of a "span" or "a" then we need to give it an exclusive parent of its own.
		// This would only happen when part of a text node is selected either at the beginning or end of a range or both.
		if((textContainer.childNodes.length > 1) &&
		   (textContainer.isXHTMLNamed("span") || textContainer.isXHTMLNamed("a")))
		{
				var siblingHolder;
				// leave any nodes before or after this one with their own copy of the container
				if(textNodes[i].previousSibling) {
					siblingHolder = textContainer.cloneNode(false);
					textContainer.parentNode.insertBefore(siblingHolder, textContainer);
					siblingHolder.appendChild(textNodes[i].previousSibling);
				}

				if(textNodes[i].nextSibling) {
					siblingHolder = textContainer.cloneNode(false);
					if(textContainer.nextSibling)
						textContainer.parentNode.insertBefore(siblingHolder, textContainer.nextSibling);
					else
						textContainer.parentNode.appendChild(siblingHolder);
					siblingHolder.appendChild(textNodes[i].nextSibling);
				}
		}

		// From now on, we assume that text has an exclusive "span" or "a" parent OR it is inside a container
		// that can have elementName inserted INTO it and around the text.
		if(textContainer.isXHTMLNamed("span")) {
			//If "span" replace "span" with "elementName".
			textContainer = textContainer.parentNode.replaceChildOnlyNS(textContainer, eDOMXHTML.NS, elementName);
		} else if (textContainer.isXHTMLNamed("a")) {
			//Else if "a" wrap "a" with elementName
			var newParent = documentCreateXHTMLElementNS(namespace, elementName);
			textContainer.parentNode.insertBefore(newParent, textContainer);
			newParent.appendChild(textContainer);
			textContainer = newParent;
		} else {
			//Else insert elementName inside container.
			var linkHolder = documentCreateXHTMLElementNS(namespace, elementName);
			textContainer.insertBefore(linkHolder, textNodes[i]);
			linkHolder.appendChild(textNodes[i]);
			textContainer = linkHolder;		
		}
		textNodes[i] = textContainer.firstChild;
	}

	// normalize elements [may be a waste - why not normalizeElements at the node level?]
	var normalizeRange = document.createRange();
	normalizeRange.selectNode(this.commonAncestorContainer);
	normalizeRange.normalizeElements(elementName);
	normalizeRange.normalizeElements("span");
	normalizeRange.normalizeElements("a");
	normalizeRange.detach();

	// now normalize text
	this.commonAncestorContainer.parentElement.normalize();
	this.__restoreTextBoundaries();
}

/**
Applies a link to this text range.

Works on the principle that there aren't nested "span" tags. It is illegal to nest "a" tags.

@param {String} href url for link.
@param {String} title hover title for link. (Optional)
*/
Range.prototype.linkText = function(href, title)
{
 try{
	// go through all text nodes in the range and link to them unless already set to this link
	var textNodes = this.textNodes();
	for(var i=0; i<textNodes.length; i++)
	{
		var textContainer = textNodes[i].parentNode;

		// If the selected text is part of a "span" or "a" we need to give it an exclusive parent of its own.
		if((textContainer.childNodes.length > 1) &&
		   (textContainer.isXHTMLNamed("span") || textContainer.isXHTMLNamed("a")))
		{ 
				var siblingHolder;
				// leave any nodes before or after this one with their own copy of the container
				if(textNodes[i].previousSibling) {
					siblingHolder = textContainer.cloneNode(false);
					textContainer.parentNode.insertBefore(siblingHolder, textContainer);
					siblingHolder.appendChild(textNodes[i].previousSibling);	
				}
				if(textNodes[i].nextSibling) {
					siblingHolder = textContainer.cloneNode(false);
					if(textContainer.nextSibling)
						textContainer.parentNode.insertBefore(siblingHolder, textContainer.nextSibling);
					else 
						textContainer.parentNode.appendChild(siblingHolder);
					siblingHolder.appendChild(textNodes[i].nextSibling);	
				}									
		}

		// From now on, we assume that text has an exclusive "a" or "span" parent OR it is inside a container
		// that can have an "a" inserted into it and around the text.
		if(!textContainer.isXHTMLNamed("a")) {
			if(textContainer.isXHTMLNamed("span")) {
				//If "span" replace "span" with "a".
				textContainer = textContainer.parentNode.replaceChildOnlyNS(textContainer, eDOMXHTML.NS, "a");
			} else {
				//Else insert "a" inside the container.
				var linkHolder = eDOMXHTML.createElement('a');
				textContainer.insertBefore(linkHolder, textNodes[i]);
				linkHolder.appendChild(textNodes[i]);
				textContainer = linkHolder;		
			}
		}
		textNodes[i] = textContainer.firstChild;
		textContainer.setAttribute("href", href);

		if ( title !== null && title != "" ) {
			textContainer.setAttribute("title", title);
		}		
		else if (textContainer.hasAttribute("title")) {
			textContainer.removeAttribute("title");
		}
	}

	// normalize elements [may be a waste - why not normalizeElements at the node level?]
	var normalizeRange = document.createRange();
	normalizeRange.selectNode(this.commonAncestorContainer);
	normalizeRange.normalizeElements("a");
	normalizeRange.normalizeElements("span");
	normalizeRange.detach();

	// now normalize text  
	this.commonAncestorContainer.parentElement.normalize();
	this.__restoreTextBoundaries();
 }catch(e){
 }
}

/**
Clear XHTML links from this CSS Text range.

It is XHTML specific.
All links in the selected range are cleared.

To Do:
Should localName for "a" be used rather than nodeName in case there is an xhtml namespace prefix?

Doesn't handle nested tags within the link.
*/
Range.prototype.clearTextLinks = function(title)
{
 try{
	// go through all text nodes in the range and clear them.
	var textNodes = this.textNodes();
	for(var i=0; i<textNodes.length; i++) {		
		// figure out this and then it's on to efficiency before subroutines ... ex of sub ... 
		// try text nodes returning one node ie/ node itself! could cut down on normalize calls ...
		var textContainer = textNodes[i].parentNode;
		if(textContainer.isXHTMLNamed("a")) {
			if(textContainer.childNodes.length > 1) {
				var siblingHolder;
				// leave any nodes before or after this one with their own copy of the container
				if(textNodes[i].previousSibling) {
					siblingHolder = textContainer.cloneNode(false);
					textContainer.parentNode.insertBefore(siblingHolder, textContainer);
					siblingHolder.appendChild(textNodes[i].previousSibling);	
				}
				if(textNodes[i].nextSibling) {
					siblingHolder = textContainer.cloneNode(false);
					if(textContainer.nextSibling)
						textContainer.parentNode.insertBefore(siblingHolder, textContainer.nextSibling);
					else 
						textContainer.parentNode.appendChild(siblingHolder);
					siblingHolder.appendChild(textNodes[i].nextSibling);	
				}
			}
			if (textContainer.hasAttribute("href"))
			//If href attribute is present remove it.
				textContainer.removeAttribute("href");
			if (title && textContainer.hasAttribute("title"))
				textContainer.removeAttribute("title");
			if(textContainer.attributes.length > 0) {
			//If any remaining attributes rename "a" to "span".
				textContainer = textContainer.parentNode.replaceChildOnlyNS(textContainer, eDOMXHTML.NS, "span");
			}else{
			//Else delete "a" node, keeping the children of course.
				textContainer.parentNode.removeChildOnly(textContainer);
			}
		}
	}

	// normalize elements. 
	var normalizeRange = document.createRange();
	normalizeRange.selectNode(this.commonAncestorContainer);
	normalizeRange.normalizeElements("a");
	normalizeRange.normalizeElements("span");
	normalizeRange.detach();

	// now normalize text
	this.commonAncestorContainer.parentElement.normalize();
	this.__restoreTextBoundaries();
 }catch(e){
 }
}

/******************************************* Misc ***********************************************/

/**
eDOMXHTML: normalize XHTML Text Style.

<p>Cleanup standard conformant inline CSS i.e. remove redundancies.
<p>Three types of "a" or "span" need normalization:
<ul>
<li>those that only contain space: delete the element but keep the space
<li>those with redundant inline CSS, CSS that their parent's have: delete the redundant settings
<li>those that follow an "a" or "span" that has identical properties: merge 
<li>those with no CSS: delete the element but keep its contents
</ul>
<p>POST04:<br />
 - change to take inline container list so that it is more generic.
*/ 
Element.prototype.__normalizeXHTMLTextStyle = function()
{
	// normalize element filter checks two things:
	// - is the element within the Range
	// - does the element have an identical previous sibling (prelude to merge!)

	var normalizeFilter = function(node) {	
		// Will only normalize "span" or "a" elements with one child.
		// This prevents bad XHTML from complicating merge logic below.	
		if( (node.parentNode.nodeNamed("span") || node.parentNode.nodeNamed("a"))
		 && (node.parentNode.childNodes.length == 1)) {
			return NodeFilter.FILTER_ACCEPT;
		}	  	
		return NodeFilter.FILTER_REJECT;		
	}

	var tw = document.createTreeWalker(this, NodeFilter.SHOW_TEXT, normalizeFilter, false);
	// go through all text nodes that appear in spans or A's
	var nextTextNode = tw.firstChild();
	while(nextTextNode)
	{
		var thisTextNode = nextTextNode;
		nextTextNode = tw.nextNode();
		var thisISC = thisTextNode.parentNode;
		thisISC.removeRedundantInlineStyles();	

		// nix isc if empty text node or redundant inline styles
		if((__NodeFilter.nonEmptyText(thisTextNode) == NodeFilter.FILTER_REJECT) ||
		   !thisISC.hasAttributes())
		{
			var elContents = document.createRange();
			elContents.selectNodeContents(thisISC);
			thisISC.parentNode.insertBefore(elContents.extractContents(), thisISC);
			thisISC.parentNode.normalize(); // no problem to tree traversal as it gets text in own spans!
			thisISC.parentNode.removeChild(thisISC);
		}
		else // if didn't nix then try to merge!
		{
			// merge ISC with previous sibling if they match
			var previousSibling = thisISC.__editablePreviousSibling;

			if(previousSibling && ("match" in previousSibling) && previousSibling.match(thisISC))
			{
				// delete intermediate text nodes and other "useless" markup
				var thisISCParentNode = thisISC.parentNode;
				while(previousSibling.nextSibling != thisISC)
					thisISCParentNode.removeChild(previousSibling.nextSibling);
	
				// merge siblings
				var contentsToMerge = document.createRange();
				contentsToMerge.selectNodeContents(thisISC);
				previousSibling.appendChild(contentsToMerge.extractContents());
				previousSibling.normalize();
				thisISCParentNode.removeChild(thisISC);	
			}
		}
	}
}

/**
eDOMXHTML: Remove redundant inline styles.
Remove redundant inline styles - those styles inherited already from the parent of an element

There can be a problem here because some style values can have different values but mean the same thing.
E.g. for font-weight "normal" and "400" may mean the same thing.
Hence, in your MES you should use the computedStyleValue rather than an equivalent alternative.

POST04: rename to normalizeStyleSettings?
Issue of default "transparent" for color and "start" for "text-align".
*/
Element.prototype.removeRedundantInlineStyles = function()
{
	for(var j=0; j<this.style.length; j++) {
		var styleName = this.style.item(j);
		var styleValue = this.style.getPropertyValue(styleName);
		var parentStyleValue = this.parentNode.computedStyleValue(styleName);

		// problem: if the element itself/class sets the style even when we remove the inline then there's a prob!

		if( eDOM.compareStyleValues(styleName, parentStyleValue, styleValue) ) {
		// if parent already has setting then remove it
			this.style.removeProperty(styleName);	

			// now make sure that removing the property actually expose's the parent's setting. If not
			// then we must put it back! It is possible that the tag of the element fixes the style or
			// at least a style different than the one intended.
			if( !eDOM.compareStyleValues( styleName, parentStyleValue, this.computedStyleValue(styleName)) )
					this.style.setProperty(styleName, styleValue, "");		
		}
	}

	if(this.hasXHTMLAttribute("style") && (this.style.length == 0))
		this.attributes.removeNamedItem("style");
}

/**
eDOMXHTML: Removes redundant CSS Style for this element.

To Do: Combine with removeRedundantInlineStyles.
*/
Element.prototype.removeRedundantStyles = function() {
	for(var j=0; j<this.style.length; j++) {
		var styleName = this.style.item(j);
		var styleValue = this.style.getPropertyValue(styleName);
		var computedStyleValue = this.computedStyleValue(styleName);
		if( eDOM.compareStyleValues(styleName, computedStyleValue, styleValue) ) {
			this.style.removeProperty(styleName);	
			if( !eDOM.compareStyleValues(styleName, computedStyleValue, this.computedStyleValue(styleName)) )
				this.style.setProperty(styleName, styleValue, "");		
		}
	}
	if(this.hasXHTMLAttribute("style") && (this.style.length == 0))
		this.attributes.removeNamedItem("style");
}


/**
Compares <code>value2</code> with <code>value</code> for Style <code>name</code>.
Always returns true if <code>value2="inherit"</code>.
Returns true if <code>value2</code> is equal to, or synonymous with, <code>value</code>.

@param {String} name : style name
@param {String} value : style value
@param {String} value2 : alternative style value
@return true if <code>value2</code> is equal to, or synonymous with, <code>value</code>.
@type boolean 
*/
EDOM.prototype.compareStyleValues = function( name, value, value2 ) {

	//rgb values
	if ( name == "color" ) {
		// rgb(255,00,00) - remove text, whitespace and brackets. Parse as numbers.
		var re = /\s|rgb|\(|\)/gi;
		var c1 = value.replace( re, "" );
		var c2 = value2.replace( re, "" );
		c1 = c1.split( "," , 3 );
		c2 = c2.split( "," , 3 );
		if ( c1.length==3 && c2.length==3
			&& parseInt(c1[0])==parseInt(c2[0])
			&& parseInt(c1[1])==parseInt(c2[1])
			&& parseInt(c1[2])==parseInt(c2[2]) ) {
			return true;
		} else {
			return false;
		}
	}

	if ( value2 == value )
		return true;
	if ( value2 == "inherit" )
		return true;
	
	var i,s;
	for (i=0; i<STYLE_SYNONYMS.length; i++) {
		s = STYLE_SYNONYMS[i];
		if ( name == s[0] ) {
			if ((value==s[1] && value2==s[2]) || (value==s[2] && value2==s[1])) return true;
		}
	}
	return false;
}
/*
Note: font-weight relative values (bolder, etc) do not correspond to absolute numbers.
Hence, there is not a comparable computed style value for these values.
E.g. if parent has font-weight of 400, and this is made "bolder", the computed style value
is returned as 401. This is a moz specific return value.
The browser may map this to 500, 600, 700, etc depending on the fonts available to the browser
from the client's computer.
In particular, the default style for the bold tag <b> is "bolder", not "bold".
I.e. you must either redefine <b> to be font-weight: bold in your style sheet,
or redefine the MES to work with relative font-weights. In general, a single MES command
is not going to work with both relative and absolute values.
*/
const STYLE_SYNONYMS = new Array(
["font-weight","bold","700"],
["font-weight","normal","400"],
["text-align","left","start"]
);


/**
eDOMXHTML: Create XHTML Table.

@param {number} noRows
@param {number} noColumns
*/
Document.prototype.createXHTMLTable = documentCreateXHTMLTable;

function documentCreateXHTMLTable(noRows, noColumns)
{
	if((/\D+/.test(noRows)) || (/\D+/.test(noColumns)) || (noRows==0) || (noColumns==0))
		return null; // go to exception

	var te = eDOMXHTML.createElement("table");
	for(var i=0; i<noRows; i++)
	{
		var tre = eDOMXHTML.createElement("tr");
		te.appendChild(tre);
		for(var j=0; j<noColumns;j++)
		{
			var tde = eDOMXHTML.createElement("td");
			tde.appendChild(document.createTextNode(STRING_NBSP)); // POST05: CHANGE TO BR TOKEN 
			tre.appendChild(tde);
		}
	}
	return te;
}

/**
Can newChild be appended to this element?
XHTML only!
There is a similar DOM 3 function that may replace this one day.
This only does a basic nesting check, expand as required.

@param {Element} newChild
@return true if allowed; false not allowed; null if don't know; Exception if newChild invalid.
@type boolean
*/
Element.prototype.canAppendChild = function( newChild ) {
	if ( newChild.nodeType == Node.ELEMENT_NODE ) {
		var childName = newChild.localName.toLowerCase();
		var parentName = this.localName.toLowerCase();

		for ( var i=0; i<MOZILE_XHTML_ELEMENTS.length; i++ ) {
			for ( var j=0; j<MOZILE_XHTML_ELEMENTS[i][0].length; j++ ) {
				if( parentName == MOZILE_XHTML_ELEMENTS[i][0][j] ) {
					for ( var k=0; k<MOZILE_XHTML_ELEMENTS[i][1].length; k++ ) {
						if ( childName == MOZILE_XHTML_ELEMENTS[i][1][k] ) {
							return true; //Allowed!
						}
					}
					//Not allowed as a child.
					return false;
				}
			}
		}
		//Not found in table.
		return null;
	} else { //Error
		throw( "canAppendChild: newChild not an element: "+newChild );
	}
}
/*
 Basic nesting - doesn't cover all possibilities.
 Derived from the xhtml 1 strict DTD.
*/
const MOZILE_XHTML_ELEMENTS =
[
/* %Flow */
[["div","li","td","th","dd","ins","del"],
 ["p","h1","h2","h3","h4","h5","h6","div","ul","ol","dl","pre","hr","blockquote","address",
 "fieldset","table","form",
 "a","br","span","bdo","object","img","map","tt","i","b","big","small",
 "em","strong","dfn","code","q","sub","sup","samp","kbd","var","cite","abbr","acronym",
 "input","select","textarea","label","button",
 "ins","del","script","noscript"]
],
/* %Block */
[["body","blockquote"],
 ["p","h1","h2","h3","h4","h5","h6","div","ul","ol","dl","pre","hr","blockquote","address",
 "fieldset","table","form",
 "ins","del","script","noscript"]
],
/* %Inline */
[["p","h1","h2","h3","h4","h5","h6","span","caption"],
 ["a","br","span","bdo","object","img","map","tt","i","b","big","small",
 "em","strong","dfn","code","q","sub","sup","samp","kbd","var","cite","abbr","acronym",
 "input","select","textarea","label","button",
 "ins","del","script","noscript"]
],
/* a */
[["a"],
 ["br","span","bdo","object","img","map","tt","i","b","big","small",
 "em","strong","dfn","code","q","sub","sup","samp","kbd","var","cite","abbr","acronym",
 "input","select","textarea","label","button",
 "ins","del","script","noscript"]
],
/* pre */
[["pre"], /* %Inline, excluding "img|object|big|small|sub|sup" */
 ["a","br","span","bdo","map","tt","i","b",
 "em","strong","dfn","code","q","samp","kbd","var","cite","abbr","acronym",
 "input","select","textarea","label","button",
 "ins","del","script","noscript"]
],
[["ol","ul"],["li"]],
[["dl"],["dt","dd"]],
/* table */
[["table"],["caption","col","colgroup","thead","tfoot","tbody","tr"]],
[["tr"],["th","td"]],
[["thead","tfoot","tbody"],["tr"]],
[["colgroup"],["col"]],
/* %Inline (2nd batch) */
[["em","strong","code","sub","sup","tt","i","b","big","small","dt","address","bdo","dfn","samp","kbd","var","cite","abbr","acronym","q"],
 ["a","br","span","bdo","object","img","map","tt","i","b","big","small",
 "em","strong","dfn","code","q","sub","sup","samp","kbd","var","cite","abbr","acronym",
 "input","select","textarea","label","button",
 "ins","del","script","noscript"]
],

];
